@@ -59,6 +59,14 @@
 #include <process.h>
 #include <direct.h>
 
+#ifdef __MINGW32__
+/* Mingw doesn't have prototypes for these */
+int _wspawnvpe (int, const wchar_t *, const wchar_t **, const wchar_t **);
+int _wspawnvp (int, const wchar_t *, const wchar_t **);
+int _wspawnve (int, const wchar_t *, const wchar_t **, const wchar_t **);
+int _wspawnv (int, const wchar_t *, const wchar_t **);
+#endif
+
 #include "glibintl.h"
 
 #ifdef G_SPAWN_WIN32_DEBUG
@@ -100,6 +108,60 @@ enum {
   ARG_COUNT = ARG_PROGRAM
 };
 
+static gchar *
+protect_argv_string (const gchar *string)
+{
+  const gchar *p = string;
+  gchar *retval, *q;
+  gint len = 0;
+  gboolean need_dblquotes = FALSE;
+  while (*p)
+    {
+      if (*p == ' ' || *p == '\t')
+	need_dblquotes = TRUE;
+      else if (*p == '"')
+	len++;
+      else if (*p == '\\')
+	{
+	  const gchar *pp = p;
+	  while (*pp && *pp == '\\')
+	    pp++;
+	  if (*pp == '"')
+	    len++;
+	}
+      len++;
+      p++;
+    }
+  
+  q = retval = g_malloc (len + need_dblquotes*2 + 1);
+  p = string;
+
+  if (need_dblquotes)
+    *q++ = '"';
+  
+  while (*p)
+    {
+      if (*p == '"')
+	*q++ = '\\';
+      else if (*p == '\\')
+	{
+	  const gchar *pp = p;
+	  while (*pp && *pp == '\\')
+	    pp++;
+	  if (*pp == '"')
+	    *q++ = '\\';
+	}
+      *q++ = *p;
+      p++;
+    }
+  
+  if (need_dblquotes)
+    *q++ = '"';
+  *q++ = '\0';
+
+  return retval;
+}
+
 static gint
 protect_argv (gchar  **argv,
 	      gchar ***new_argv)
@@ -115,7 +177,8 @@ protect_argv (gchar  **argv,
    * reconstructed correctly in the C runtime startup code.  Note that
    * the unquoting algorithm in the C runtime is really weird, and
    * rather different than what Unix shells do. See stdargv.c in the C
-   * runtime sources (in the Platform SDK, in src/crt).
+   * runtime sources (in the Platform SDK, in src/crt), and
+   * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/vccelng/htm/progs_12.asp
    *
    * Note that an new_argv[0] constructed by this function should
    * *not* be passed as the filename argument to a spawn* or exec*
@@ -123,56 +186,8 @@ protect_argv (gchar  **argv,
    * without any quoting.
    */
   for (i = 0; i < argc; i++)
-    {
-      gchar *p = argv[i];
-      gchar *q;
-      gint len = 0;
-      gboolean need_dblquotes = FALSE;
-      while (*p)
-	{
-	  if (*p == ' ' || *p == '\t')
-	    need_dblquotes = TRUE;
-	  else if (*p == '"')
-	    len++;
-	  else if (*p == '\\')
-	    {
-	      gchar *pp = p;
-	      while (*pp && *pp == '\\')
-		pp++;
-	      if (*pp == '"')
-		len++;
-	    }
-	  len++;
-	  p++;
-	}
+    (*new_argv)[i] = protect_argv_string (argv[i]);
 
-      q = (*new_argv)[i] = g_malloc (len + need_dblquotes*2 + 1);
-      p = argv[i];
-
-      if (need_dblquotes)
-	*q++ = '"';
-
-      while (*p)
-	{
-	  if (*p == '"')
-	    *q++ = '\\';
-	  else if (*p == '\\')
-	    {
-	      gchar *pp = p;
-	      while (*pp && *pp == '\\')
-		pp++;
-	      if (*pp == '"')
-		*q++ = '\\';
-	    }
-	  *q++ = *p;
-	  p++;
-	}
-
-      if (need_dblquotes)
-	*q++ = '"';
-      *q++ = '\0';
-      /* printf ("argv[%d]:%s, need_dblquotes:%s len:%d => %s\n", i, argv[i], need_dblquotes?"TRUE":"FALSE", len, (*new_argv)[i]); */
-    }
   (*new_argv)[argc] = NULL;
 
   return argc;
@@ -192,25 +207,25 @@ g_spawn_error_quark (void)
 }
 
 gboolean
-g_spawn_async (const gchar          *working_directory,
-               gchar               **argv,
-               gchar               **envp,
-               GSpawnFlags           flags,
-               GSpawnChildSetupFunc  child_setup,
-               gpointer              user_data,
-               GPid                 *child_handle,
-               GError              **error)
+g_spawn_async_utf8 (const gchar          *working_directory,
+		    gchar               **argv,
+		    gchar               **envp,
+		    GSpawnFlags           flags,
+		    GSpawnChildSetupFunc  child_setup,
+		    gpointer              user_data,
+		    GPid                 *child_handle,
+		    GError              **error)
 {
   g_return_val_if_fail (argv != NULL, FALSE);
   
-  return g_spawn_async_with_pipes (working_directory,
-                                   argv, envp,
-                                   flags,
-                                   child_setup,
-                                   user_data,
-                                   child_handle,
-                                   NULL, NULL, NULL,
-                                   error);
+  return g_spawn_async_with_pipes_utf8 (working_directory,
+					argv, envp,
+					flags,
+					child_setup,
+					user_data,
+					child_handle,
+					NULL, NULL, NULL,
+					error);
 }
 
 /* Avoids a danger in threaded situations (calling close()
@@ -258,9 +273,7 @@ read_data (GString     *str,
     goto again;
   else if (giostatus == G_IO_STATUS_ERROR)
     {
-      g_set_error (error,
-                   G_SPAWN_ERROR,
-                   G_SPAWN_ERROR_READ,
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_READ,
                    _("Failed to read data from child process"));
       
       return READ_FAILED;
@@ -275,9 +288,7 @@ make_pipe (gint     p[2],
 {
   if (pipe (p) < 0)
     {
-      g_set_error (error,
-                   G_SPAWN_ERROR,
-                   G_SPAWN_ERROR_FAILED,
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
                    _("Failed to create pipe for communicating with child process (%s)"),
                    g_strerror (errno));
       return FALSE;
@@ -315,9 +326,7 @@ read_helper_report (int      fd,
         {
           /* Some weird shit happened, bail out */
               
-          g_set_error (error,
-                       G_SPAWN_ERROR,
-                       G_SPAWN_ERROR_FAILED,
+          g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
                        _("Failed to read from child pipe (%s)"),
                        g_strerror (errno));
 
@@ -343,17 +352,13 @@ set_child_error (gint         report[2],
   switch (report[0])
     {
     case CHILD_CHDIR_FAILED:
-      g_set_error (error,
-		   G_SPAWN_ERROR,
-		   G_SPAWN_ERROR_CHDIR,
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_CHDIR,
 		   _("Failed to change to directory '%s' (%s)"),
 		   working_directory,
 		   g_strerror (report[1]));
       break;
     case CHILD_SPAWN_FAILED:
-      g_set_error (error,
-		   G_SPAWN_ERROR,
-		   G_SPAWN_ERROR_FAILED,
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
 		   _("Failed to execute child process (%s)"),
 		   g_strerror (report[1]));
       break;
@@ -363,8 +368,242 @@ set_child_error (gint         report[2],
 }
 
 static gboolean
-do_spawn_with_pipes (gboolean              dont_wait,
-		     gboolean		   dont_return_handle,
+utf8_charv_to_wcharv (char     **utf8_charv,
+		      wchar_t ***wcharv,
+		      int       *error_index,
+		      GError   **error)
+{
+  wchar_t **retval = NULL;
+
+  *wcharv = NULL;
+  if (utf8_charv != NULL)
+    {
+      int n = 0, i;
+
+      while (utf8_charv[n])
+	n++;
+      retval = g_new (wchar_t *, n + 1);
+
+      for (i = 0; i < n; i++)
+	{
+	  retval[i] = g_utf8_to_utf16 (utf8_charv[i], -1, NULL, NULL, error);
+	  if (retval[i] == NULL)
+	    {
+	      if (error_index)
+		*error_index = i;
+	      while (i)
+		g_free (retval[--i]);
+	      g_free (retval);
+	      return FALSE;
+	    }
+	}
+	    
+      retval[n] = NULL;
+    }
+  *wcharv = retval;
+  return TRUE;
+}
+
+static gboolean
+utf8_charv_to_cp_charv (char   **utf8_charv,
+			gchar ***cp_charv,
+			int     *error_index,
+			GError **error)
+{
+  char **retval = NULL;
+
+  *cp_charv = NULL;
+  if (utf8_charv != NULL)
+    {
+      int n = 0, i;
+
+      while (utf8_charv[n])
+	n++;
+      retval = g_new (char *, n + 1);
+
+      for (i = 0; i < n; i++)
+	{
+	  retval[i] = g_locale_from_utf8 (utf8_charv[i], -1, NULL, NULL, error);
+	  if (retval[i] == NULL)
+	    {
+	      if (error_index)
+		*error_index = i;
+	      while (i)
+		g_free (retval[--i]);
+	      g_free (retval);
+	      return FALSE;
+	    }
+	}
+      retval[n] = NULL;
+    }
+
+  *cp_charv = retval;
+  return TRUE;
+}
+
+static gboolean
+do_spawn_directly (gint                 *exit_status,
+		   gboolean		 do_return_handle,
+		   GSpawnFlags           flags,
+		   gchar               **argv,
+		   char                **envp,
+		   char                **protected_argv,
+		   GSpawnChildSetupFunc  child_setup,
+		   gpointer              user_data,
+		   GPid                 *child_handle,
+		   GError              **error)     
+{
+  const int mode = (exit_status == NULL) ? P_NOWAIT : P_WAIT;
+  char **new_argv;
+  int rc = -1;
+  int saved_errno;
+  GError *conv_error = NULL;
+  gint conv_error_index;
+
+  new_argv = (flags & G_SPAWN_FILE_AND_ARGV_ZERO) ? protected_argv + 1 : protected_argv;
+  if (G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      wchar_t *wargv0, **wargv, **wenvp;
+      
+      wargv0 = g_utf8_to_utf16 (argv[0], -1, NULL, NULL, &conv_error);
+      if (wargv0 == NULL)
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid program name: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  
+	  return FALSE;
+	}
+      
+      if (!utf8_charv_to_wcharv (new_argv, &wargv, &conv_error_index, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in argument vector at %d: %s",
+		       conv_error_index, conv_error->message);
+	  g_error_free (conv_error);
+	  g_free (wargv0);
+	  
+	  return FALSE;
+	}
+      
+      if (!utf8_charv_to_wcharv (envp, &wenvp, NULL, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in environment: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  g_free (wargv0);
+	  g_strfreev ((gchar **) wargv);
+	  
+	  return FALSE;
+	}
+      
+      if (child_setup)
+	(* child_setup) (user_data);
+      
+      if (flags & G_SPAWN_SEARCH_PATH)
+	if (wenvp != NULL)
+	  rc = _wspawnvpe (mode, wargv0, (const wchar_t **) wargv, (const wchar_t **) wenvp);
+	else
+	  rc = _wspawnvp (mode, wargv0, (const wchar_t **) wargv);
+      else
+	if (wenvp != NULL)
+	  rc = _wspawnve (mode, wargv0, (const wchar_t **) wargv, (const wchar_t **) wenvp);
+	else
+	  rc = _wspawnv (mode, wargv0, (const wchar_t **) wargv);
+      
+      g_free (wargv0);
+      g_strfreev ((gchar **) wargv);
+      g_strfreev ((gchar **) wenvp);
+    }
+  else
+    {
+      char *cpargv0, **cpargv, **cpenvp;
+      
+      cpargv0 = g_locale_from_utf8 (argv[0], -1, NULL, NULL, &conv_error);
+      if (cpargv0 == NULL)
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid program name: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+
+	  return FALSE;
+	}
+
+      if  (!utf8_charv_to_cp_charv (new_argv, &cpargv, &conv_error_index, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in argument vector at %d: %s",
+		       conv_error_index, conv_error->message);
+	  g_error_free (conv_error);
+	  g_free (cpargv0);
+
+	  return FALSE;
+	}
+
+      if (!utf8_charv_to_cp_charv (envp, &cpenvp, NULL, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in environment: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  g_free (cpargv0);
+	  g_strfreev (cpargv);
+
+	  return FALSE;
+	}
+
+      if (child_setup)
+	(* child_setup) (user_data);
+
+      if (flags & G_SPAWN_SEARCH_PATH)
+	if (cpenvp != NULL)
+	  rc = spawnvpe (mode, cpargv0, (const char **) cpargv, (const char **) cpenvp);
+	else
+	  rc = spawnvp (mode, cpargv0, (const char **) cpargv);
+      else
+	if (envp != NULL)
+	  rc = spawnve (mode, cpargv0, (const char **) cpargv, (const char **) cpenvp);
+	else
+	  rc = spawnv (mode, cpargv0, (const char **) cpargv);
+
+      g_free (cpargv0);
+      g_strfreev (cpargv);
+      g_strfreev (cpenvp);
+    }
+
+  saved_errno = errno;
+
+  if (rc == -1 && saved_errno != 0)
+    {
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		   _("Failed to execute child process (%s)"),
+		   g_strerror (saved_errno));
+      return FALSE;
+    }
+
+  if (exit_status == NULL)
+    {
+      if (child_handle && do_return_handle)
+	*child_handle = (GPid) rc;
+      else
+	{
+	  CloseHandle ((HANDLE) rc);
+	  if (child_handle)
+	    *child_handle = 0;
+	}
+    }
+  else
+    *exit_status = rc;
+
+  return TRUE;
+}
+
+static gboolean
+do_spawn_with_pipes (gint                 *exit_status,
+		     gboolean		   do_return_handle,
 		     const gchar          *working_directory,
 		     gchar               **argv,
 		     char                **envp,
@@ -375,7 +614,6 @@ do_spawn_with_pipes (gboolean              dont_wait,
 		     gint                 *standard_input,
 		     gint                 *standard_output,
 		     gint                 *standard_error,
-		     gint                 *exit_status,
 		     gint		  *err_report,
 		     GError              **error)     
 {
@@ -391,9 +629,20 @@ do_spawn_with_pipes (gboolean              dont_wait,
   int stderr_pipe[2] = { -1, -1 };
   int child_err_report_pipe[2] = { -1, -1 };
   int helper_report[2];
+  static gboolean warned_about_child_setup = FALSE;
+  GError *conv_error = NULL;
+  gint conv_error_index;
+  gchar *helper_process;
+  CONSOLE_CURSOR_INFO cursor_info;
   
   SETUP_DEBUG();
 
+  if (child_setup && !warned_about_child_setup)
+    {
+      warned_about_child_setup = TRUE;
+      g_warning ("passing a child setup function to the g_spawn functions is pointless and dangerous on Win32");
+    }
+
   argc = protect_argv (argv, &protected_argv);
 
   if (!standard_input && !standard_output && !standard_error &&
@@ -404,65 +653,15 @@ do_spawn_with_pipes (gboolean              dont_wait,
       (flags & G_SPAWN_LEAVE_DESCRIPTORS_OPEN))
     {
       /* We can do without the helper process */
-      int mode = dont_wait ? P_NOWAIT : P_WAIT;
-
-      if (debug)
-	g_print ("doing without " HELPER_PROCESS "\n");
-
-      /* Call user function just before we spawn the program. Dunno
-       * what's the usefulness of this. A child setup function used on
-       * Unix probably isn't of much use as such on Win32, anyhow.
-       */
-      if (child_setup)
-	(* child_setup) (user_data);
-
-      new_argv = (flags & G_SPAWN_FILE_AND_ARGV_ZERO) ? protected_argv + 1 : protected_argv;
-      if (flags & G_SPAWN_SEARCH_PATH)
-	if (envp != NULL)
-	  rc = spawnvpe (mode, argv[0], (const char **) new_argv, (const char **) envp);
-	else
-	  rc = spawnvp (mode, argv[0], (const char **) new_argv);
-      else
-	if (envp != NULL)
-	  rc = spawnve (mode, argv[0], (const char **) new_argv, (const char **) envp);
-	else
-	  rc = spawnv (mode, argv[0], (const char **) new_argv);
-
-      saved_errno = errno;
-
-      for (i = 0; i < argc; i++)
-	g_free (protected_argv[i]);
-      g_free (protected_argv);
-
-      if (rc == -1 && saved_errno != 0)
-	{
-	  g_set_error (error,
-		       G_SPAWN_ERROR,
-		       G_SPAWN_ERROR_FAILED,
-		       _("Failed to execute child process (%s)"),
-		       g_strerror (errno));
-	  goto cleanup_and_fail;
-	}
-
-      if (dont_wait)
-	{
-	  if (child_handle && !dont_return_handle)
-	    *child_handle = (GPid) rc;
-	  else
-	    {
-	      CloseHandle (rc);
-	      if (child_handle)
-		*child_handle = 0;
-	    }
-	}
-      else if (exit_status)
-	*exit_status = rc;
-      
-      return TRUE;
+      gboolean retval =
+	do_spawn_directly (exit_status, do_return_handle, flags,
+			   argv, envp, protected_argv,
+			   child_setup, user_data, child_handle,
+			   error);
+      g_strfreev (protected_argv);
+      return retval;
     }
 
-  new_argv = g_new (char *, argc + 1 + ARG_COUNT);
-
   if (standard_input && !make_pipe (stdin_pipe, error))
     goto cleanup_and_fail;
   
@@ -475,7 +674,12 @@ do_spawn_with_pipes (gboolean              dont_wait,
   if (!make_pipe (child_err_report_pipe, error))
     goto cleanup_and_fail;
   
-  new_argv[0] = HELPER_PROCESS;
+  new_argv = g_new (char *, argc + 1 + ARG_COUNT);
+  if (GetConsoleCursorInfo (GetStdHandle (STD_OUTPUT_HANDLE), &cursor_info))
+    helper_process = HELPER_PROCESS "-console.exe";
+  else
+    helper_process = HELPER_PROCESS ".exe";
+  new_argv[0] = helper_process;
   _g_sprintf (args[ARG_CHILD_ERR_REPORT], "%d", child_err_report_pipe[1]);
   new_argv[ARG_CHILD_ERR_REPORT] = args[ARG_CHILD_ERR_REPORT];
   
@@ -532,8 +736,7 @@ do_spawn_with_pipes (gboolean              dont_wait,
     }
   
   if (working_directory && *working_directory)
-    /* The g_strdup() to lose the constness */
-    new_argv[ARG_WORKING_DIRECTORY] = g_strdup (working_directory);
+    new_argv[ARG_WORKING_DIRECTORY] = protect_argv_string (working_directory);
   else
     new_argv[ARG_WORKING_DIRECTORY] = g_strdup ("-");
   
@@ -547,7 +750,7 @@ do_spawn_with_pipes (gboolean              dont_wait,
   else
     new_argv[ARG_USE_PATH] = "-";
 
-  if (dont_wait)
+  if (exit_status == NULL)
     new_argv[ARG_WAIT] = "-";
   else
     new_argv[ARG_WAIT] = "w";
@@ -557,35 +760,126 @@ do_spawn_with_pipes (gboolean              dont_wait,
 
   if (debug)
     {
-      g_print ("calling " HELPER_PROCESS " with argv:\n");
+      g_print ("calling %s with argv:\n", helper_process);
       for (i = 0; i < argc + 1 + ARG_COUNT; i++)
 	g_print ("argv[%d]: %s\n", i, (new_argv[i] ? new_argv[i] : "NULL"));
     }
 
-  if (child_setup)
-    (* child_setup) (user_data);
+  if (G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      wchar_t *whelper = g_utf8_to_utf16 (helper_process, -1, NULL, NULL, NULL);
+      wchar_t **wargv, **wenvp;
 
-  if (envp != NULL)
-    /* Let's hope envp hasn't mucked with PATH so that
-     * gspawn-win32-helper.exe isn't found.
-     */
-    rc = spawnvpe (P_NOWAIT, HELPER_PROCESS, (const char **) new_argv, (const char **) envp);
+      if (!utf8_charv_to_wcharv (new_argv, &wargv, &conv_error_index, &conv_error))
+	{
+	  if (conv_error_index == ARG_WORKING_DIRECTORY)
+	    g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_CHDIR,
+			 "Invalid working directory: %s",
+			 conv_error->message);
+	  else
+	    g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+			 "Invalid string in argument vector at %d: %s",
+			 conv_error_index - ARG_PROGRAM, conv_error->message);
+	  g_error_free (conv_error);
+	  g_strfreev (protected_argv);
+	  g_free (new_argv[ARG_WORKING_DIRECTORY]);
+	  g_free (new_argv);
+	  g_free (whelper);
+	  
+	  return FALSE;
+	}
+      
+      if (!utf8_charv_to_wcharv (envp, &wenvp, NULL, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in environment: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  g_strfreev (protected_argv);
+	  g_free (new_argv[ARG_WORKING_DIRECTORY]);
+	  g_free (new_argv);
+	  g_free (whelper);
+	  g_strfreev ((gchar **) wargv);
+	  
+	  return FALSE;
+	}
+
+      if (child_setup)
+	(* child_setup) (user_data);
+
+      if (wenvp != NULL)
+	/* Let's hope envp hasn't mucked with PATH so that
+	 * gspawn-win32-helper.exe isn't found.
+	 */
+	rc = _wspawnvpe (P_NOWAIT, whelper, (const wchar_t **) wargv, (const wchar_t **) wenvp);
+      else
+	rc = _wspawnvp (P_NOWAIT, whelper, (const wchar_t **) wargv);
+
+      saved_errno = errno;
+
+      g_free (whelper);
+      g_strfreev ((gchar **) wargv);
+      g_strfreev ((gchar **) wenvp);
+    }
   else
-    rc = spawnvp (P_NOWAIT, HELPER_PROCESS, (const char **) new_argv);
+    {
+      char **cpargv, **cpenvp;
 
-  saved_errno = errno;
+      if (!utf8_charv_to_cp_charv (new_argv, &cpargv, &conv_error_index, &conv_error))
+	{
+	  if (conv_error_index == ARG_WORKING_DIRECTORY)
+	    g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_CHDIR,
+			 "Invalid working directory: %s",
+			 conv_error->message);
+	  else
+	    g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+			 "Invalid string in argument vector at %d: %s",
+			 conv_error_index - ARG_PROGRAM, conv_error->message);
+	  g_error_free (conv_error);
+	  g_strfreev (protected_argv);
+	  g_free (new_argv[ARG_WORKING_DIRECTORY]);
+	  g_free (new_argv);
+	  
+	  return FALSE;
+	}
+	
+      if (!utf8_charv_to_cp_charv (envp, &cpenvp, NULL, &conv_error))
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in environment: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  g_strfreev (protected_argv);
+	  g_free (new_argv[ARG_WORKING_DIRECTORY]);
+	  g_free (new_argv);
+	  g_strfreev (cpargv);
+	  
+	  return FALSE;
+	}
 
-  /* Close thed the other process's ends of the pipes in this
-   * process, otherwise the reader will never get EOF.
+      if (child_setup)
+	(* child_setup) (user_data);
+
+      if (cpenvp != NULL)
+	rc = spawnvpe (P_NOWAIT, helper_process, (const char **) cpargv, (const char **) cpenvp);
+      else
+	rc = spawnvp (P_NOWAIT, helper_process, (const char **) cpargv);
+
+      saved_errno = errno;
+
+      g_strfreev (cpargv);
+      g_strfreev (cpenvp);
+    }
+
+  /* Close the other process's ends of the pipes in this process,
+   * otherwise the reader will never get EOF.
    */
   close_and_invalidate (&child_err_report_pipe[1]);
   close_and_invalidate (&stdin_pipe[0]);
   close_and_invalidate (&stdout_pipe[1]);
   close_and_invalidate (&stderr_pipe[1]);
 
-  for (i = 0; i < argc; i++)
-    g_free (protected_argv[i]);
-  g_free (protected_argv);
+  g_strfreev (protected_argv);
 
   g_free (new_argv[ARG_WORKING_DIRECTORY]);
   g_free (new_argv);
@@ -593,14 +887,13 @@ do_spawn_with_pipes (gboolean              dont_wait,
   /* Check if gspawn-win32-helper couldn't be run */
   if (rc == -1 && saved_errno != 0)
     {
-      g_set_error (error,
-		   G_SPAWN_ERROR,
-		   G_SPAWN_ERROR_FAILED,
-		   _("Failed to execute helper program"));
+      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		   _("Failed to execute helper program (%s)"),
+		   g_strerror (saved_errno));
       goto cleanup_and_fail;
     }
 
-  if (!dont_wait)
+  if (exit_status != NULL)
     {
       /* Synchronous case. Pass helper's report pipe back to caller,
        * which takes care of reading it after the grandchild has
@@ -620,7 +913,7 @@ do_spawn_with_pipes (gboolean              dont_wait,
       switch (helper_report[0])
 	{
 	case CHILD_NO_ERROR:
-	  if (child_handle && dont_wait && !dont_return_handle)
+	  if (child_handle && do_return_handle)
 	    {
 	      /* rc is our HANDLE for gspawn-win32-helper. It has
 	       * told us the HANDLE of its child. Duplicate that into
@@ -633,8 +926,6 @@ do_spawn_with_pipes (gboolean              dont_wait,
 	    }
 	  else if (child_handle)
 	    *child_handle = 0;
-	  if (exit_status)
-	    *exit_status = helper_report[1];
 	  break;
 	  
 	default:
@@ -680,16 +971,16 @@ do_spawn_with_pipes (gboolean              dont_wait,
 }
 
 gboolean
-g_spawn_sync (const gchar          *working_directory,
-              gchar               **argv,
-              gchar               **envp,
-              GSpawnFlags           flags,
-              GSpawnChildSetupFunc  child_setup,
-              gpointer              user_data,
-              gchar               **standard_output,
-              gchar               **standard_error,
-              gint                 *exit_status,
-              GError              **error)     
+g_spawn_sync_utf8 (const gchar          *working_directory,
+		   gchar               **argv,
+		   gchar               **envp,
+		   GSpawnFlags           flags,
+		   GSpawnChildSetupFunc  child_setup,
+		   gpointer              user_data,
+		   gchar               **standard_output,
+		   gchar               **standard_error,
+		   gint                 *exit_status,
+		   GError              **error)     
 {
   gint outpipe = -1;
   gint errpipe = -1;
@@ -723,8 +1014,8 @@ g_spawn_sync (const gchar          *working_directory,
   if (standard_error)
     *standard_error = NULL;
   
-  if (!do_spawn_with_pipes (FALSE,
-			    TRUE,
+  if (!do_spawn_with_pipes (&status,
+			    FALSE,
 			    working_directory,
 			    argv,
 			    envp,
@@ -735,7 +1026,6 @@ g_spawn_sync (const gchar          *working_directory,
 			    NULL,
 			    standard_output ? &outpipe : NULL,
 			    standard_error ? &errpipe : NULL,
-			    &status,
 			    &reportpipe,
 			    error))
     return FALSE;
@@ -749,9 +1039,12 @@ g_spawn_sync (const gchar          *working_directory,
       outstr = g_string_new (NULL);
       outchannel = g_io_channel_win32_new_fd (outpipe);
       g_io_channel_set_encoding (outchannel, NULL, NULL);
+      g_io_channel_set_buffered (outchannel, FALSE);
       g_io_channel_win32_make_pollfd (outchannel,
 				      G_IO_IN | G_IO_ERR | G_IO_HUP,
 				      &outfd);
+      if (debug)
+	g_print ("outfd=%x\n", outfd.fd);
     }
       
   if (errpipe >= 0)
@@ -759,9 +1052,12 @@ g_spawn_sync (const gchar          *working_directory,
       errstr = g_string_new (NULL);
       errchannel = g_io_channel_win32_new_fd (errpipe);
       g_io_channel_set_encoding (errchannel, NULL, NULL);
+      g_io_channel_set_buffered (errchannel, FALSE);
       g_io_channel_win32_make_pollfd (errchannel,
 				      G_IO_IN | G_IO_ERR | G_IO_HUP,
 				      &errfd);
+      if (debug)
+	g_print ("errfd=%x\n", errfd.fd);
     }
 
   /* Read data until we get EOF on all pipes. */
@@ -791,9 +1087,7 @@ g_spawn_sync (const gchar          *working_directory,
         {
           failed = TRUE;
 
-          g_set_error (error,
-                       G_SPAWN_ERROR,
-                       G_SPAWN_ERROR_READ,
+          g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_READ,
                        _("Unexpected error in g_io_channel_win32_poll() reading data from a child process"));
 	  
           break;
@@ -919,6 +1213,225 @@ g_spawn_sync (const gchar          *working_directory,
     }
 }
 
+gboolean
+g_spawn_async_with_pipes_utf8 (const gchar          *working_directory,
+			       gchar               **argv,
+			       gchar               **envp,
+			       GSpawnFlags           flags,
+			       GSpawnChildSetupFunc  child_setup,
+			       gpointer              user_data,
+			       GPid                 *child_handle,
+			       gint                 *standard_input,
+			       gint                 *standard_output,
+			       gint                 *standard_error,
+			       GError              **error)
+{
+  g_return_val_if_fail (argv != NULL, FALSE);
+  g_return_val_if_fail (standard_output == NULL ||
+                        !(flags & G_SPAWN_STDOUT_TO_DEV_NULL), FALSE);
+  g_return_val_if_fail (standard_error == NULL ||
+                        !(flags & G_SPAWN_STDERR_TO_DEV_NULL), FALSE);
+  /* can't inherit stdin if we have an input pipe. */
+  g_return_val_if_fail (standard_input == NULL ||
+                        !(flags & G_SPAWN_CHILD_INHERITS_STDIN), FALSE);
+  
+  return do_spawn_with_pipes (NULL,
+			      (flags & G_SPAWN_DO_NOT_REAP_CHILD),
+			      working_directory,
+			      argv,
+			      envp,
+			      flags,
+			      child_setup,
+			      user_data,
+			      child_handle,
+			      standard_input,
+			      standard_output,
+			      standard_error,
+			      NULL,
+			      error);
+}
+
+gboolean
+g_spawn_command_line_sync_utf8 (const gchar  *command_line,
+				gchar       **standard_output,
+				gchar       **standard_error,
+				gint         *exit_status,
+				GError      **error)
+{
+  gboolean retval;
+  gchar **argv = 0;
+
+  g_return_val_if_fail (command_line != NULL, FALSE);
+  
+  if (!g_shell_parse_argv (command_line,
+                           NULL, &argv,
+                           error))
+    return FALSE;
+  
+  retval = g_spawn_sync_utf8 (NULL,
+			      argv,
+			      NULL,
+			      G_SPAWN_SEARCH_PATH,
+			      NULL,
+			      NULL,
+			      standard_output,
+			      standard_error,
+			      exit_status,
+			      error);
+  g_strfreev (argv);
+
+  return retval;
+}
+
+gboolean
+g_spawn_command_line_async_utf8 (const gchar *command_line,
+				 GError     **error)
+{
+  gboolean retval;
+  gchar **argv = 0;
+
+  g_return_val_if_fail (command_line != NULL, FALSE);
+
+  if (!g_shell_parse_argv (command_line,
+                           NULL, &argv,
+                           error))
+    return FALSE;
+  
+  retval = g_spawn_async_utf8 (NULL,
+			       argv,
+			       NULL,
+			       G_SPAWN_SEARCH_PATH,
+			       NULL,
+			       NULL,
+			       NULL,
+			       error);
+  g_strfreev (argv);
+
+  return retval;
+}
+
+void
+g_spawn_close_pid (GPid pid)
+{
+    CloseHandle (pid);
+}
+
+/* Binary compatibility versions that take system codepage pathnames,
+ * argument vectors and environments. These get used only by code
+ * built against 2.8.1 or earlier. Code built against 2.8.2 or later
+ * will use the _utf8 versions above (see the #defines in gspawn.h).
+ */
+
+#undef g_spawn_async
+#undef g_spawn_async_with_pipes
+#undef g_spawn_sync
+#undef g_spawn_command_line_sync
+#undef g_spawn_command_line_async
+
+static gboolean
+setup_utf8_copies (const gchar *working_directory,
+		   gchar      **utf8_working_directory,
+		   gchar      **argv,
+		   gchar     ***utf8_argv,
+		   gchar      **envp,
+		   gchar     ***utf8_envp,
+		   GError     **error)
+{
+  gint i, argc, envc;
+
+  if (working_directory == NULL)
+    *utf8_working_directory = NULL;
+  else
+    {
+      GError *conv_error = NULL;
+      
+      *utf8_working_directory = g_locale_to_utf8 (working_directory, -1, NULL, NULL, &conv_error);
+      if (*utf8_working_directory == NULL)
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_CHDIR,
+		       "Invalid working directory: %s",
+		       conv_error->message);
+	  g_error_free (conv_error);
+	  return FALSE;
+	}
+    }
+
+  argc = 0;
+  while (argv[argc])
+    ++argc;
+  *utf8_argv = g_new (gchar *, argc + 1);
+  for (i = 0; i < argc; i++)
+    {
+      GError *conv_error = NULL;
+
+      (*utf8_argv)[i] = g_locale_to_utf8 (argv[i], -1, NULL, NULL, &conv_error);
+      if ((*utf8_argv)[i] == NULL)
+	{
+	  g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+		       "Invalid string in argument vector at %d: %s",
+		       i, conv_error->message);
+	  g_error_free (conv_error);
+	  
+	  g_strfreev (*utf8_argv);
+	  *utf8_argv = NULL;
+
+	  g_free (*utf8_working_directory);
+	  *utf8_working_directory = NULL;
+
+	  return FALSE;
+	}
+    }
+  (*utf8_argv)[argc] = NULL;
+
+  if (envp == NULL)
+    {
+      *utf8_envp = NULL;
+    }
+  else
+    {
+      envc = 0;
+      while (envp[envc])
+	++envc;
+      *utf8_envp = g_new (gchar *, envc + 1);
+      for (i = 0; i < envc; i++)
+	{
+	  GError *conv_error = NULL;
+
+	  (*utf8_envp)[i] = g_locale_to_utf8 (envp[i], -1, NULL, NULL, &conv_error);
+	  if ((*utf8_envp)[i] == NULL)
+	    {	
+	      g_set_error (error, G_SPAWN_ERROR, G_SPAWN_ERROR_FAILED,
+			   "Invalid string in environment: %s",
+			   conv_error->message);
+	      g_error_free (conv_error);
+
+	      g_strfreev (*utf8_envp);
+	      *utf8_envp = NULL;
+
+	      g_strfreev (*utf8_argv);
+	      *utf8_argv = NULL;
+
+	      g_free (*utf8_working_directory);
+	      *utf8_working_directory = NULL;
+
+	      return FALSE;
+	    }
+	}
+      (*utf8_envp)[envc] = NULL;
+    }
+  return TRUE;
+}
+
+static void
+free_utf8_copies (gchar  *utf8_working_directory,
+		  gchar **utf8_argv,
+		  gchar **utf8_envp)
+{
+  g_free (utf8_working_directory);
+  g_strfreev (utf8_argv);
+  g_strfreev (utf8_envp);
+}
+
 gboolean
 g_spawn_async_with_pipes (const gchar          *working_directory,
                           gchar               **argv,
@@ -932,38 +1445,89 @@ g_spawn_async_with_pipes (const gchar          *working_directory,
                           gint                 *standard_error,
                           GError              **error)
 {
-  g_return_val_if_fail (argv != NULL, FALSE);
-  g_return_val_if_fail (standard_output == NULL ||
-                        !(flags & G_SPAWN_STDOUT_TO_DEV_NULL), FALSE);
-  g_return_val_if_fail (standard_error == NULL ||
-                        !(flags & G_SPAWN_STDERR_TO_DEV_NULL), FALSE);
-  /* can't inherit stdin if we have an input pipe. */
-  g_return_val_if_fail (standard_input == NULL ||
-                        !(flags & G_SPAWN_CHILD_INHERITS_STDIN), FALSE);
-  
-  return do_spawn_with_pipes (TRUE,
-			      !(flags & G_SPAWN_DO_NOT_REAP_CHILD),
-			      working_directory,
-			      argv,
-			      envp,
-			      flags,
-			      child_setup,
-			      user_data,
-			      child_handle,
-			      standard_input,
-			      standard_output,
-			      standard_error,
-			      NULL,
-			      NULL,
+  gchar *utf8_working_directory;
+  gchar **utf8_argv;
+  gchar **utf8_envp;
+  gboolean retval;
+
+  if (!setup_utf8_copies (working_directory, &utf8_working_directory,
+			  argv, &utf8_argv,
+			  envp, &utf8_envp,
+			  error))
+    return FALSE;
+
+  retval = g_spawn_async_with_pipes_utf8 (utf8_working_directory,
+					  utf8_argv, utf8_envp,
+					  flags, child_setup, user_data,
+					  child_handle,
+					  standard_input, standard_output, standard_error,
+					  error);
+
+  free_utf8_copies (utf8_working_directory, utf8_argv, utf8_envp);
+
+  return retval;
+}
+
+gboolean
+g_spawn_async (const gchar          *working_directory,
+	       gchar               **argv,
+	       gchar               **envp,
+	       GSpawnFlags           flags,
+	       GSpawnChildSetupFunc  child_setup,
+	       gpointer              user_data,
+	       GPid                 *child_handle,
+	       GError              **error)
+{
+  return g_spawn_async_with_pipes (working_directory,
+				   argv, envp,
+				   flags,
+				   child_setup,
+				   user_data,
+				   child_handle,
+				   NULL, NULL, NULL,
+				   error);
+}
+
+gboolean
+g_spawn_sync (const gchar          *working_directory,
+	      gchar               **argv,
+	      gchar               **envp,
+	      GSpawnFlags           flags,
+	      GSpawnChildSetupFunc  child_setup,
+	      gpointer              user_data,
+	      gchar               **standard_output,
+	      gchar               **standard_error,
+	      gint                 *exit_status,
+	      GError              **error)     
+{
+  gchar *utf8_working_directory;
+  gchar **utf8_argv;
+  gchar **utf8_envp;
+  gboolean retval;
+
+  if (!setup_utf8_copies (working_directory, &utf8_working_directory,
+			  argv, &utf8_argv,
+			  envp, &utf8_envp,
+			  error))
+    return FALSE;
+
+  retval = g_spawn_sync_utf8 (utf8_working_directory,
+			      utf8_argv, utf8_envp,
+			      flags, child_setup, user_data,
+			      standard_output, standard_error, exit_status,
 			      error);
+
+  free_utf8_copies (utf8_working_directory, utf8_argv, utf8_envp);
+
+  return retval;
 }
 
 gboolean
 g_spawn_command_line_sync (const gchar  *command_line,
-                           gchar       **standard_output,
-                           gchar       **standard_error,
-                           gint         *exit_status,
-                           GError      **error)
+			   gchar       **standard_output,
+			   gchar       **standard_error,
+			   gint         *exit_status,
+			   GError      **error)
 {
   gboolean retval;
   gchar **argv = 0;
@@ -992,7 +1556,7 @@ g_spawn_command_line_sync (const gchar  *command_line,
 
 gboolean
 g_spawn_command_line_async (const gchar *command_line,
-                            GError     **error)
+			    GError     **error)
 {
   gboolean retval;
   gchar **argv = 0;
@@ -1019,11 +1583,5 @@ g_spawn_command_line_async (const gchar *command_line,
 
 #endif /* !GSPAWN_HELPER */
 
-void
-g_spawn_close_pid (GPid pid)
-{
-    CloseHandle (pid);
-}
-
 #define __G_SPAWN_C__
 #include "galiasdef.c"
